#pragma once
#include <math.h>

/******************************************************************************\
\******************************************************************************/

template<class T, int N> class Vector;
template<class T> class Vector2;
template<class T> class Vector3;
template<class T> class Vector4;

// Some common shorthands
typedef Vector2<float>							Vec2;
typedef Vector3<float>							Vec3U;
typedef	Vector3<float> __declspec(align(16))	Vec3A;
typedef	Vector4<float> __declspec(align(16))	Vec4;

typedef Vec3U									Vec3;

typedef Vector<int, 2>							IVec2;
typedef Vector<char, 4> __declspec(align(4))	CVec4;

template<class T, int N> class Vector
{
	public:

	// Constructors
	Vector()							{}
	inline Vector(const Vector& v)		{ for (int i = 0; i < N; i++) Set(i) = v.Get(i); }
	inline Vector(const T& s)			{ for (int i = 0; i < N; i++) Set(i) = s; }
	inline Vector(const Vector<T, N-1>& v, const T& s)	{ for (int i = 0; i < N-1; i++) Set(i) = v.Get(i); Set(N-1) = s; }
	inline Vector(const T& s, const Vector<T, N-1>& v)	{ Set(0) = s; for (int i = 1; i < N; i++) Set(i) = v.Get(i-1); }

	// Accessors
	inline const T&	Get(const int index) const			{ return _elem[index]; }
	inline T&		Set(const int index)				{ return _elem[index]; }

	// Arithmetic
	inline friend	const T			operator|(const Vector& a, const Vector& b)
		{ float r = (T)0;	for (int i = 0; i < N; i++) r += a.Get(i) * b.Get(i);	return r; }
	inline friend	const Vector	operator+(const Vector& a, const Vector& b)
		{ Vector r;			for (int i = 0; i < N; i++) r.Set(i) = a.Get(i) + b.Get(i);	return r; }
	inline friend	const Vector	operator-(const Vector& a, const Vector& b)
		{ Vector r;			for (int i = 0; i < N; i++) r.Set(i) = a.Get(i) - b.Get(i);	return r; }
	inline friend	const Vector	operator*(const Vector& a, const Vector& b)
		{ Vector r;			for (int i = 0; i < N; i++) r.Set(i) = a.Get(i) * b.Get(i);	return r; }
	inline friend	const Vector	operator/(const Vector& a, const Vector& b)
		{ Vector r;			for (int i = 0; i < N; i++) r.Set(i) = a.Get(i) / b.Get(i);	return r; }
	inline friend	const Vector	operator-(const Vector& a)
		{ Vector r;			for (int i = 0; i < N; i++) r.Set(i) = -a.Get(i);			return r; }
	inline friend	const T			LengthSquared(const Vector& a)				{ return a | a; }
	inline friend	const T			Length(const Vector& a)						{ return (T)sqrt(LengthSquared(a)); }
	inline friend	const Vector	Normalize(const Vector& a)					{ return a / Length(a); }

	// Comparison
	inline friend	const bool		operator==(const Vector& a, const Vector& b)
		{ bool r = true;	for (int i = 0; i < N; i++) r = r && a.Get(i) == b.Get(i);		return r; }

	// Permutation
	inline friend	const Vector	operator>>(const Vector& a, const int count)
		{ Vector r;			for (int i = 0; i < N; i++) r.Set(i) = a.Get((i+N-count) % N);	return r; }

	// Assignment
	inline void operator=(const Vector& b)		{ for (int i = 0; i < N; i++) Set(i) = b.Get(i); }
	inline void operator+=(const Vector& b)		{ for (int i = 0; i < N; i++) Set(i) += b.Get(i); }
	inline void operator-=(const Vector& b)		{ for (int i = 0; i < N; i++) Set(i) -= b.Get(i); }
	inline void operator*=(const Vector& b)		{ for (int i = 0; i < N; i++) Set(i) *= b.Get(i); }
	inline void operator/=(const Vector& b)		{ for (int i = 0; i < N; i++) Set(i) /= b.Get(i); }

	protected:

	T		_elem[N];
};

template<class T> class Vector2:	public Vector<T, 2>
{
	public:
	inline Vector2()	{}
	inline Vector2(const Vector<T, 2>& v): Vector<T, 2>(v)	{}
	inline Vector2(const T& s): Vector<T, 2>(s)				{}
	inline Vector2(const T& x, const T& y)	{ Set(0) = x; Set(1) = y; }

	// Cross product
	inline friend	const T	operator^(const Vector2& a, const Vector2& b)
		{ return a.Get(0)*b.Get(1) - a.Get(1)*b.Get(0); }
};

template<class T> class Vector3:	public Vector<T, 3>
{
	public:
	inline Vector3()	{}
	inline Vector3(const Vector<T, 3>& v): Vector<T, 3>(v)	{}
	inline Vector3(const Vector<T, 2>& v, const T& s): Vector<T, 3>(v, s)	{}
	inline Vector3(const T& s, const Vector<T, 2>& v): Vector<T, 3>(s, v)	{}
	inline Vector3(const T& s): Vector<T, 3>(s)				{}
	inline Vector3(const T& x, const T& y, const T& z)		{ Set(0) = x; Set(1) = y; Set(2) = z; }

	// Cross product
	inline friend	const Vector3	operator^(const Vector3& a, const Vector3& b)
		{ return Vector3(a.Get(1)*b.Get(2) - a.Get(2)*b.Get(1), a.Get(2)*b.Get(0) - a.Get(0)*b.Get(2), a.Get(0)*b.Get(1) - a.Get(1)*b.Get(0)); }
};

template<class T> class Vector4:	public Vector<T, 4>
{
	public:
	inline Vector4()	{}
	inline Vector4(const Vector<T, 4>& v): Vector<T, 4>(v)	{}
	inline Vector4(const Vector<T, 3>& v, const T& s): Vector<T, 4>(v, s)	{}
	inline Vector4(const T& s, const Vector<T, 3>& v): Vector<T, 4>(s, v)	{}
	inline Vector4(const float& s): Vector<T, 4>(s)				{}
	inline Vector4(const T& x, const T& y, const T& z, const T& w)	{ Set(0) = x; Set(1) = y; Set(2) = z; Set(3) = w; }
};

/******************************************************************************\
\******************************************************************************/
